@@ -2,7 +2,7 @@
     Plain implementation of the plainest possible MonteCarlo
 """
 
-from vegasflow.configflow import DTYPE, fone, fzero
+from vegasflow.configflow import fone, fzero
 from vegasflow.monte_carlo import MonteCarloFlow, wrapper, sampler
 import tensorflow as tf
 
@@ -12,6 +12,8 @@ class PlainFlow(MonteCarloFlow):
     Simple Monte Carlo integrator.
     """
 
+    _CAN_RUN_VECTORIAL = True
+
     def _run_event(self, integrand, ncalls=None):
         if ncalls is None:
             n_events = self.n_events
@@ -19,13 +21,16 @@ class PlainFlow(MonteCarloFlow):
             n_events = ncalls
 
         # Generate all random number for this iteration
-        rnds, _, xjac = self._generate_random_array(n_events)
+        rnds, xjac = self._generate_random_array(n_events)
+
         # Compute the integrand
-        tmp = integrand(rnds, n_dim=self.n_dim, weight=xjac) * xjac
+        tmp = integrand(rnds, weight=xjac) * xjac
         tmp2 = tf.square(tmp)
-        # Accumulate the current result
-        res = tf.reduce_sum(tmp)
-        res2 = tf.reduce_sum(tmp2)
+
+        # Accommodate multidimensional output by ensuring that only the event axis is accumulated
+        res = tf.reduce_sum(tmp, axis=0)
+        res2 = tf.reduce_sum(tmp2, axis=0)
+
         return res, res2
 
     def _run_iteration(self):
@@ -38,9 +43,10 @@ class PlainFlow(MonteCarloFlow):
 
 
 def plain_wrapper(*args, **kwargs):
-    """ Wrapper around PlainFlow """
+    """Wrapper around PlainFlow"""
     return wrapper(PlainFlow, *args, **kwargs)
 
+
 def plain_sampler(*args, **kwargs):
-    """ Wrapper sampler around PlainFlow """
+    """Wrapper sampler around PlainFlow"""
     return sampler(PlainFlow, *args, **kwargs)
