@@ -1,16 +1,24 @@
-#!/usr/env python
 """
-    This module contains the VegasFlow class and all its auxuliary functions
+    This module contains the VegasFlow class and all its auxiliary functions
 
     The main interfaces of this class are the class `VegasFlow` and the
     `vegas_wrapper`
 """
-from vegasflow.configflow import DTYPE, DTYPEINT, fone, fzero, float_me, ione, int_me
 import json
 import numpy as np
 import tensorflow as tf
 
-from vegasflow.configflow import BINS_MAX, ALPHA
+from vegasflow.configflow import (
+    DTYPE,
+    DTYPEINT,
+    fone,
+    fzero,
+    float_me,
+    ione,
+    int_me,
+    BINS_MAX,
+    ALPHA,
+)
 from vegasflow.monte_carlo import MonteCarloFlow, wrapper, sampler
 from vegasflow.utils import consume_array_into_indices
 
@@ -20,8 +28,66 @@ logger = logging.getLogger(__name__)
 
 FBINS = float_me(BINS_MAX)
 
+
+@tf.function(
+    input_signature=[
+        tf.TensorSpec(shape=[None, None], dtype=DTYPE),
+        tf.TensorSpec(shape=[None, BINS_MAX + 1], dtype=DTYPE),
+    ]
+)
+def importance_sampling_digest(xn, divisions):
+    """Importance sampling algorithm:
+    receives a random array (number of dimensions, number of dim)
+    containing information about from which bins in the
+    grid (n_dims, BINS_MAX+1) the random points have to be sampled
+
+    This algorithm is shared between the simplest form of Vegas
+    (VegasFlow: only importance sampling)
+    and Vegas+ (VegasFlowPlus: importance and stratified sampling)
+    and so it has been lifted to its own function
+
+    Parameters:
+    ----------
+        xn: float tensor (n_dim, n_events)
+            which bins to sample from
+        divisions: float tensor (n_dims, BINS_MAX+1)
+            grid of divisions for the importance sampling algorithm
+
+    Returns
+    -------
+        ind_i: integer tensor (n_events, n_dim)
+            index in the divisions grid from which the points should be sampled
+        x: float tensor (n_events, n_dim)
+            random values sampled in the divisions grid
+        xdelta: float tensor (n_events,)
+            weight of the random points
+    """
+    ind_i = int_me(xn)
+    # Get the value of the left and right sides of the bins
+    ind_f = ind_i + ione
+    x_ini = tf.gather(divisions, ind_i, batch_dims=1)
+    x_fin = tf.gather(divisions, ind_f, batch_dims=1)
+    # Compute the width of the bins
+    xdelta = x_fin - x_ini
+    # Take the decimal part of bin (i.e., how deep within the bin)
+    aux_rand = xn - tf.math.floor(xn)
+    x = x_ini + xdelta * aux_rand
+    # Compute the weight of the points
+    weights = tf.reduce_prod(xdelta * FBINS, axis=0)
+
+    # Tranpose the output to be what the external functions expect
+    x = tf.transpose(x)
+    ind_i = tf.transpose(ind_i)
+    return ind_i, x, weights
+
+
 # Auxiliary functions for Vegas
-@tf.function
+@tf.function(
+    input_signature=[
+        tf.TensorSpec(shape=[None, None], dtype=DTYPE),
+        tf.TensorSpec(shape=[None, BINS_MAX + 1], dtype=DTYPE),
+    ]
+)
 def _generate_random_array(rnds, divisions):
     """
     Generates the Vegas random array for any number of events
@@ -30,17 +96,17 @@ def _generate_random_array(rnds, divisions):
     ----------
         rnds: array shaped (None, n_dim)
             Random numbers used as an input for Vegas
-        divisions: array shaped (n_dim, BINS_MAX)
+        divisions: array shaped (n_dim, BINS_MAX+1)
             vegas grid
 
     Returns
     -------
         x: array (None, n_dim)
             Vegas random output
-        div_index: array (None, n_dim)
-            division index in which each (n_dim) set of random numbers fall
         w: array (None,)
             Weight of each set of (n_dim) random numbers
+        div_index: array (None, n_dim)
+            division index in which each (n_dim) set of random numbers fall
     """
     # Get the boundaries of the random numbers
     #     reg_i = fzero
@@ -48,36 +114,22 @@ def _generate_random_array(rnds, divisions):
     # Get the index of the division we are interested in
     xn = FBINS * (fone - tf.transpose(rnds))
 
-    @tf.function
-    def digest(xn):
-        ind_i = tf.cast(xn, DTYPEINT)
-        # Get the value of the left and right sides of the bins
-        ind_f = ind_i + ione
-        x_ini = tf.gather(divisions, ind_i, batch_dims=1)
-        x_fin = tf.gather(divisions, ind_f, batch_dims=1)
-        # Compute the width of the bins
-        xdelta = x_fin - x_ini
-        return ind_i, x_ini, xdelta
-
-    ind_i, x_ini, xdelta = digest(xn)
     # Compute the random number between 0 and 1
-    # This is the heavy part of the calc
-    @tf.function
-    def compute_x(x_ini, xn, xdelta):
-        aux_rand = xn - tf.math.floor(xn)
-        return x_ini + xdelta * aux_rand
+    # and the index of the bin where it has been sampled from
+    ind_xn, x, weights = importance_sampling_digest(xn, divisions)
 
-    x = compute_x(x_ini, xn, xdelta)
     # Compute the random number between the limits
+    # commented, for now only from 0 to 1
     #     x = reg_i + rand_x * (reg_f - reg_i)
-    # and the weight
-    weights = tf.reduce_prod(xdelta * FBINS, axis=0)
-    x_t = tf.transpose(x)
-    int_xn = tf.transpose(ind_i)
-    return x_t, int_xn, weights
+    return x, weights, ind_xn
 
 
-@tf.function
+@tf.function(
+    input_signature=[
+        tf.TensorSpec(shape=[BINS_MAX], dtype=DTYPE),
+        tf.TensorSpec(shape=[BINS_MAX + 1], dtype=DTYPE),
+    ]
+)
 def refine_grid_per_dimension(t_res_sq, subdivisions):
     """
     Modifies the boundaries for the vegas grid for a given dimension
@@ -95,17 +147,8 @@ def refine_grid_per_dimension(t_res_sq, subdivisions):
             array with the new boundaries of the grid
     """
     # Define some constants
-    paddings = tf.constant(
-        [
-            [1, 1],
-        ]
-    )
-    tmp_meaner = tf.fill(
-        [
-            BINS_MAX - 2,
-        ],
-        float_me(3.0),
-    )
+    paddings = int_me([[1, 1]])
+    tmp_meaner = tf.fill([BINS_MAX - 2], float_me(3.0))
     meaner = tf.pad(tmp_meaner, paddings, constant_values=2.0)
     # Pad the vector of results
     res_padded = tf.pad(t_res_sq, paddings)
@@ -127,7 +170,14 @@ def refine_grid_per_dimension(t_res_sq, subdivisions):
         to beat the average"""
         return bin_weight < ave_t
 
-    @tf.function
+    @tf.function(
+        input_signature=[
+            tf.TensorSpec(shape=[], dtype=DTYPE),
+            tf.TensorSpec(shape=[], dtype=DTYPEINT),
+            tf.TensorSpec(shape=[], dtype=DTYPE),
+            tf.TensorSpec(shape=[], dtype=DTYPE),
+        ]
+    )
     def while_body(bin_weight, n_bin, cur, prev):
         """Fills the bin weight until it surpassed the avg
         once it's done, returns the limits of the last bin"""
@@ -165,33 +215,59 @@ def refine_grid_per_dimension(t_res_sq, subdivisions):
 ####### VegasFlow
 class VegasFlow(MonteCarloFlow):
     """
-    Implementation of the important sampling algorithm from Vegas
+    Implementation of the important sampling algorithm from Vegas.
+
+    Parameters
+    ----------
+        n_dim: int
+            number of dimensions to be integrated
+        n_events: int
+            number of events per iteration
+        train: bool
+            whether to train the grid
+        main_dimension: int
+            in case of vectorial output, main dimenison in which to train
     """
 
-    def __init__(self, n_dim, n_events, train=True, **kwargs):
+    def __init__(self, n_dim, n_events, train=True, main_dimension=0, **kwargs):
         super().__init__(n_dim, n_events, **kwargs)
 
         # If training is True, the grid will be changed after every iteration
         # otherwise it will be frozen
         self.train = train
-        self.iteration_content = None
-        self.compile_args = None
 
         # Initialize grid
         self.grid_bins = BINS_MAX + 1
         subdivision_np = np.linspace(0, 1, self.grid_bins)
         divisions_np = subdivision_np.repeat(n_dim).reshape(-1, n_dim).T
         self.divisions = tf.Variable(divisions_np, dtype=DTYPE)
+        self._main_dimension = main_dimension
+
+    def _can_run_vectorial(self, expected_shape):
+        # only implemented for the main class at the moment, not for children
+        if self._main_dimension >= expected_shape[-1]:
+            raise ValueError(
+                f"""The main dimension index ({self._main_dimension}) is greater than the dimensionality of the output ({expected_shape[-1]}).
+            Remember that arrays in python are 0-indexed!"""
+            )
+        return self.__class__.__name__ == "VegasFlow"
+
+    def make_differentiable(self):
+        """Freeze the grid if the function is to be called within a graph"""
+        if self.train:
+            logger.warning("Freezing the grid")
+            self.freeze_grid()
+        return super().make_differentiable()
 
     def freeze_grid(self):
-        """ Stops the grid from refining any more """
+        """Stops the grid from refining any more"""
         self.train = False
-        self.recompile()
+        self._recompile()
 
     def unfreeze_grid(self):
-        """ Enable the refining of the grid """
+        """Enable the refining of the grid"""
         self.train = True
-        self.recompile()
+        self._recompile()
 
     def save_grid(self, file_name):
         """Save the `divisions` array in a json file
@@ -202,8 +278,8 @@ class VegasFlow(MonteCarloFlow):
             Filename in which to save the checkpoint
         """
         div_np = self.divisions.numpy()
-        if self.integrand:
-            int_name = self.integrand.__name__
+        if self._integrand:
+            int_name = self._integrand.__name__
         else:
             int_name = ""
         json_dict = {
@@ -241,8 +317,8 @@ class VegasFlow(MonteCarloFlow):
             grid_dim = json_dict.get("dimensions")
             grid_bins = json_dict.get("BINS")
             # Check that the integrand is the same one
-            if self.integrand:
-                integrand_name = self.integrand.__name__
+            if self._integrand:
+                integrand_name = self._integrand.__name__
                 integrand_grid = json_dict.get("integrand")
                 if integrand_name != integrand_grid:
                     logger.warning(
@@ -286,12 +362,30 @@ class VegasFlow(MonteCarloFlow):
             new_divisions = refine_grid_per_dimension(arr_res2[j, :], self.divisions[j, :])
             self.divisions[j, :].assign(new_divisions)
 
-    def _generate_random_array(self, n_events):
-        """Uses the internal array to generate ``n_events`` random numbers"""
-        rnds, _, xjac = super()._generate_random_array(n_events)
-        # Pass them through the Vegas digestion
-        x, ind, w = _generate_random_array(rnds, self.divisions)
-        return x, ind, w * xjac
+    def _digest_random_generation(self, rnds):
+        """Generates ``n_events`` random numbers sampled in the
+        adapted Vegas Grid"""
+        x, w, ind = _generate_random_array(rnds, self.divisions)
+        return x, w, ind
+
+    def _importance_sampling_array_filling(self, results2, indices):
+        """Receives an array of results squared for every event
+        and an array of indices describing in which bin each result fall.
+        Fills a array with the total result in each bin to be used by
+        the importance sampling algorithm
+        """
+        if not self.train:
+            return []
+
+        arr_res2 = []
+        # If the training is active, save the result of the integral sq
+        for j in range(self.n_dim):
+            arr_res2.append(
+                consume_array_into_indices(
+                    results2, indices[:, j : j + 1], int_me(self.grid_bins - 1)
+                )
+            )
+        return tf.reshape(arr_res2, (self.n_dim, -1))
 
     def _run_event(self, integrand, ncalls=None):
         """Runs one step of Vegas.
@@ -313,31 +407,31 @@ class VegasFlow(MonteCarloFlow):
             n_events = ncalls
 
         # Generate all random number for this iteration
-        x, ind, xjac = self._generate_random_array(n_events)
+        x, xjac, ind = self._generate_random_array(n_events)
 
         # Now compute the integrand
-        if self.simplify_signature:
-            tmp = xjac * integrand(x)
-        else:
-            tmp = xjac * integrand(x, n_dim=self.n_dim, weight=xjac)
+        int_result = integrand(x, weight=xjac)
+
+        if self._vectorial:
+            xjac = tf.reshape(xjac, (-1, 1))
+        tmp = xjac * int_result
         tmp2 = tf.square(tmp)
 
         # Compute the final result for this step
-        res = tf.reduce_sum(tmp)
-        res2 = tf.reduce_sum(tmp2)
+        res = tf.reduce_sum(tmp, axis=0)
+        res2 = tf.reduce_sum(tmp2, axis=0)
 
-        arr_res2 = []
-        if self.train:
-            # If the training is active, save the result of the integral sq
-            for j in range(self.n_dim):
-                arr_res2.append(
-                    consume_array_into_indices(tmp2, ind[:, j : j + 1], int_me(self.grid_bins - 1))
-                )
-            arr_res2 = tf.reshape(arr_res2, (self.n_dim, -1))
+        # If this is a vectorial integrand, make sure that only the main dimension
+        # is used for the grid training
+        if self._vectorial:
+            tmp2 = tmp2[:, self._main_dimension]
+
+        arr_res2 = self._importance_sampling_array_filling(tmp2, ind)
 
         return res, res2, arr_res2
 
     def _iteration_content(self):
+        """Steps to follow per iteration"""
         # Compute the result
         res, res2, arr_res2 = self.run_event()
         # Compute the error
@@ -348,23 +442,9 @@ class VegasFlow(MonteCarloFlow):
             self.refine_grid(arr_res2)
         return res, sigma
 
-    def compile(self, integrand, compilable=True, **kwargs):
-        self.compile_args = (integrand, compilable, kwargs)
-        super().compile(integrand, compilable=compilable, **kwargs)
-        self.iteration_content = self._iteration_content
-
-    def recompile(self):
-        """Forces recompilation with the same arguments that have
-        previously been used for compilation"""
-        if self.compile_args is None:
-            raise RuntimeError("recompile was called without ever having called compile")
-        a = self.compile_args
-        self.compile(a[0], a[1], **a[2])
-
     def _run_iteration(self):
-        """ Runs one iteration of the Vegas integrator """
-        res, sigma = self.iteration_content()
-        return res, sigma
+        """Runs one iteration of the Vegas integrator"""
+        return self._iteration_content()
 
 
 def vegas_wrapper(integrand, n_dim, n_iter, total_n_events, **kwargs):
@@ -384,6 +464,7 @@ def vegas_wrapper(integrand, n_dim, n_iter, total_n_events, **kwargs):
     """
     return wrapper(VegasFlow, integrand, n_dim, n_iter, total_n_events, **kwargs)
 
+
 def vegas_sampler(*args, **kwargs):
     """Convenience wrapper for sampling random numbers
 
